import numpy as np

import sys
import cv2
import time
import datetime
import re
from pyzm.helpers.Base import Base
from pyzm.helpers.Media import MediaStream

import time
import requests

# Class to handle Yolo based detection




class Object(Base):

    def __init__(self, options={}, logger=None):

        Base.__init__(self,logger)
        self.model = None
        self.options = options

        if self.options.get('object_framework') == 'opencv':
            import pyzm.ml.yolo as yolo
            self.model =  yolo.Yolo(options=options, logger=logger)
            

        elif self.options.get('object_framework') == 'coral_edgetpu':
            import pyzm.ml.coral_edgetpu as tpu
            self.model = tpu.Tpu(options=options, logger=logger)

        else:
            raise ValueError ('Invalid object_framework:{}'.format(self.options.get('object_framework')))

    def get_model(self):
            return self.model

    def get_classes(self):
            return self.model.get_classes()

    def acquire_lock(self):
        self.model.acquire_lock()

    def release_lock(self):
        self.model.release_lock()

        
    def detect(self,image=None):
        h,w = image.shape[:2]
        b,l,c = self.model.detect(image)
        self.logger.Debug (2,'core model detection over, got {} objects. Now filtering'.format(len(b)))
        # Apply various object filtering rules
        max_object_area = 0
        if self.options.get('max_detection_size'):
                self.logger.Debug(3,'Max object size found to be: {}'.format(self.options.get('max_detection_size')))
                # Let's make sure its the right size
                m = re.match('(\d*\.?\d*)(px|%)?$', self.options.get('max_detection_size'),
                            re.IGNORECASE)
                if m:
                    max_object_area = float(m.group(1))
                    if m.group(2) == '%':
                        max_object_area = float(m.group(1))/100.0*(h * w)
                        self.logger.Debug (2,'Converted {}% to {}'.format(m.group(1), max_object_area))
                else:
                    self.logger.Error('max_object_area misformatted: {} - ignoring'.format(
                        self.options.get('max_object_area')))

        boxes=[]
        labels=[]
        confidences=[]

        for idx,box in enumerate(b):
            (sX,sY,eX,eY) = box
            if max_object_area:
                object_area = abs((eX-sX)*(eY-sY))
                if (object_area > max_object_area):
                    self.logger.Debug (1,'Ignoring object:{}, as it\'s area: {}px exceeds max_object_area of {}px'.format(l[idx], object_area, max_object_area))
                    continue
            if c[idx] >= self.options.get('object_min_confidence'):
                boxes.append([sX,sY,eX,eY])
                labels.append(l[idx])
                confidences.append(c[idx])
            else:
                self.logger.Debug (1,'Ignoring {} {} as conf. level {} is lower than {}'.format(l[idx],box,c[idx],self.options.get('object_min_confidence')))
       
        self.logger.Debug (2,'Returning filtered list of {} objects.'.format(len(boxes)))
        return boxes,labels,confidences



"""
utils
======
Set of utility functions
"""

from pyzm.helpers.Base import ConsoleLog
from configparser import ConfigParser

def read_config(file):
    config_file = ConfigParser(interpolation=None)
    config_file.read(file)
    return config_file

def get(key=None, section=None, conf=None):
    if conf.has_option(section, key):
        return conf.get(section, key)
    else:
        return None


import numpy as np
import sys
import time
import datetime
import re
import cv2
from pyzm.helpers.Base import Base


from PIL import Image
import portalocker
import os

from edgetpu.detection.engine import DetectionEngine

# Class to handle Yolo based detection


class Tpu(Base):

    def __init__(self, options={},logger=None ):
        Base.__init__(self,logger)
        self.classes = {}
        self.options = options
       #self.logger.Debug (1, 'UID:{} EUID:{}'.format( os.getuid(), os.geteuid()))
        
        self.processor='tpu'
        self.lock_maximum=options.get(self.processor+'_max_processes') or 1
        self.lock_name='pyzm_'+self.processor+'_lock'
        self.lock_timeout = options.get(self.processor+'_max_lock_wait') or 120

        self.logger.Debug (2,f'Semaphore: max:{self.lock_maximum}, name:{self.lock_name}, timeout:{self.lock_timeout}')
        self.lock = portalocker.BoundedSemaphore(maximum=self.lock_maximum, name=self.lock_name,timeout=self.lock_timeout)
        self.is_locked = False
        self.model = None
        self.populate_class_labels()

    def acquire_lock(self):
        if self.is_locked:
            self.logger.Debug (1, '{} Lock already acquired'.format(self.lock_name))
            return
        try:
            self.logger.Debug (1,'Waiting for TPU lock...')
            self.lock.acquire()
            self.logger.Debug (1,'Got TPU Lock')
            self.is_locked = True

        except portalocker.AlreadyLocked:
            self.logger.Error ('Timeout waiting for {} lock for {} seconds'.format(self.processor, self.lock_timeout))
            raise ValueError ('Timeout waiting for {} lock for {} seconds'.format(self.processor, self.lock_timeout))

    def release_lock(self):
        if not self.is_locked:
            self.logger.Debug (1, '{} Lock already released'.format(self.lock_name))
            return
        self.lock.release()
        self.is_locked = False
        self.logger.Debug (1,'Released TPU lock')


    def populate_class_labels(self):
        class_file_abs_path = self.options.get('object_labels')
        for row in open(class_file_abs_path):
            # unpack the row and update the labels dictionary
            (classID, label) = row.strip().split(" ", maxsplit=1)
            self.classes[int(classID)] = label.strip()

    def get_classes(self):
        return self.classes

    def load_model(self):
        self.logger.Debug (1, 'Loading TPU model from disk')
        start = datetime.datetime.now()
        self.model = DetectionEngine(self.options.get('object_weights'))
        diff_time = (datetime.datetime.now() - start).microseconds / 1000
        self.logger.Debug(
            1,'TPU initialization (loading model from disk) took: {} milliseconds'
            .format(diff_time))
        
    def detect(self, image=None):
        Height, Width = image.shape[:2]
        img = image.copy()
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(img)

        if self.options.get('auto_lock',True):
            self.acquire_lock()

        if not self.model:
           self.load_model()

        self.logger.Debug(1,
            '|---------- TPU (input image: {}w*{}h) ----------|'
            .format(Width, Height))
        start = datetime.datetime.now()
        outs = self.model.detect_with_image(img, threshold=int(self.options.get('object_min_confidence')),
                keep_aspect_ratio=True, relative_coord=False)
        diff_time = (datetime.datetime.now() - start).microseconds / 1000

        if self.options.get('auto_lock',True):
            self.release_lock()

        self.logger.Debug(
            1,'Coral TPU detection took: {} milliseconds'.format(diff_time))
      
        bbox = []
        labels = []
        conf = []

        for out in outs:
            box = out.bounding_box.flatten().astype("int")
            (startX, startY, endX, endY) = box
            bbox.append([
                    int(round(startX)),
                    int(round(startY)),
                    int(round(endX)),
                    int(round(endY))
                ])
            labels.append(self.classes[out.label_id])
            conf.append(float(out.score))

            
        return bbox, labels, conf



"""
Media
======
Generic media class that can read frames from either
a media file or via ZM APIs/index.php

"""

from pyzm.helpers.Base import Base
import requests
import cv2
import numpy as np
import imutils
from imutils.video import FileVideoStream
import time

class MediaStream(Base):
    def __init__(self, stream=None, type='video', options={}):
        self.stream = stream
        self.type = type
        self.fvs = None
        self.next_frameid_to_read = 1
        self.last_frameid_read = 0
        self.options = options
        self.more_images_to_read = True
        self.frames_processed = 0
        self.frames_read = 0
        self.api = self.options.get('api')
        self.frame_set = []
        self.next_frame_set_index = 0
        

        if self.options.get('logger'):
            self.logger = options.get('logger')
        elif self.options.get('api'):
            self.logger = options.get('api').get_logger() 

       
        self.start_frame = int(options.get('start_frame',1))
        self.frame_skip = int(options.get('frame_skip', 1))
        self.max_frames = int(options.get('max_frames', 0))

        
        if self.stream.isnumeric(): 
        # assume it is an eid, in which case we 
        # need to access it via ZM API
            if not self.options.get('api'):
                self.logger.Error('API object not provided, cannot download event {}'.format(self.stream))
                raise ValueError('API object not provided, cannot download event {}'.format(self.stream))
            
            # we can either download the event, or access it via ZM
            # indirection 
            if self.options.get('download'):
                self.logger.Debug(1,'Downloading event:{}'.format(self.stream))
                es = self.api.events({'event_id': int(self.stream)})
                if not len(es.list()):
                    self.logger.Error('No such event {} found with API'.format(self.stream))
                    return
                e = es.list()[0]
                self.stream = e.download_video(dir='/tmp')
                
                self.type = 'video'
            else:
                eid = self.stream
                self.next_frameid_to_read = int(self.start_frame)
                self.stream = '{}/index.php?view=image&width={}&eid={}'.format(self.api.get_portalbase(), self.options.get('resize', 800), eid)
                if self.api.get_auth() != '':
                    self.stream += '&{}'.format(self.api.get_auth())

                self.logger.Debug(1,'Using URL {} for stream'.format(stream))
                self.type = 'image'
        
        
        if self.options.get('frame_set'):
            self.frame_set = self.options.get('frame_set').split(',')
            if 'alarm' in self.frame_set or 'snapshot' in self.frame_set:
                if not self.api or self.type == 'video':
                    # if we are not using ZM indirection, we cannot use 'alarm' 'snapshot' etc.
                
                    self.logger.Error ('You are using frame_types that require ZM indirection')
                    raise ValueError ('You are using frame_types that require ZM indirection')
            self.max_frames = len (self.frame_set)
            self.logger.Debug (2, 'We will only process frames: {}'.format(self.frame_set))
            self.next_frame_set_index = 0
            self.start_frame = self.frame_set[self.next_frame_set_index]
    
        if (self.type=='video'):
            self.logger.Debug (1, 'Starting video stream {}'.format(stream))
            self.fvs = FileVideoStream(self.stream).start()
            time.sleep(1)
            self.logger.Debug (1, 'First load - skipping to frame {}'.format(self.start_frame))
            if self.frame_set:
                while self.fvs.more() and self.frames_read < int(self.frame_set[self.next_frame_set_index])-1:
                    self.fvs.read() 
                    self.frames_read +=1

            else:
                while self.fvs.more() and self.next_frameid_to_read < int(self.start_frame):
                    self.fvs.read()
                    self.next_frameid_to_read += 1

        else:
            self.logger.Debug (1,'No need to start streams, we are picking images from {}'.format(self.stream))


    def get_last_read_frame(self):
        return self.last_frameid_read

    def more(self):
        if (self.frame_set):
            return self.next_frame_set_index < len(self.frame_set)

        if (self.frames_processed >= self.max_frames):
            self.logger.Debug(1, 'Bailing as we have read {} frames'.format(self.max_frames))
            return False
        else:
            if (self.type == 'video'):
                return self.fvs.more()
            else:
                return self.more_images_to_read

    def stop(self):
        if (self.type == 'video'):
            self.fvs.stop()

    def read(self):
        
        if (self.type == 'video'):
            while True:
                frame = self.fvs.read()
                self.frames_read +=1

                if frame is None:
                    self.logger.Error ('Error reading frames')
                    return

                if self.frame_set and (self.frames_read != int(self.frame_set[self.next_frame_set_index])):
                    self.logger.Debug (4,'Ignoring frame {}'.format(self.frames_read))
                    continue
                else:
                    self.logger.Debug (4,'MATCHED frame {}'.format(self.frames_read))                    

                    # At this stage weare at the frame to read
                if self.frame_set:
                    self.last_frameid_read = self.frame_set[self.next_frame_set_index]
                    self.next_frame_set_index += 1
                    if self.next_frame_set_index < len (self.frame_set):
                        self.logger.Debug (4,"Now moving to frame: {}".format(self.frame_set[self.next_frame_set_index])) 

                else:
                    self.last_frameid_read = self.next_frameid_to_read
                    self.next_frameid_to_read +=1
                    if (self.last_frameid_read - 1 ) % self.frame_skip:
                        self.logger.Debug(5,'Skipping frame {}'.format(self.last_frameid_read))
                        continue    
                    
                self.logger.Debug(2, 'Processing frame:{}'.format(self.last_frameid_read))
                self.frames_processed += 1
                break
            frame = imutils.resize(frame,width=self.options.get('resize', 800))
            return frame
        
        else: # image
            if self.frame_set:
                url = '{}&fid={}'.format(self.stream,self.frame_set[self.next_frame_set_index])
            else:
                url = '{}&fid={}'.format(self.stream,self.next_frameid_to_read)
            
            self.logger.Debug (2, 'Reading {}'.format(url))

            response = requests.get(url)
            if self.frame_set:
                self.last_frameid_read = self.frame_set[self.next_frame_set_index]
                self.next_frame_set_index += 1
            else:
                self.last_frameid_read = self.next_frameid_to_read
                self.next_frameid_to_read +=self.frame_skip
            self.frames_processed +=1 

            if response.status_code == 200:
                try:
                    img = np.asarray(bytearray(response.content), dtype='uint8')
                    img = cv2.imdecode (img, cv2.IMREAD_COLOR)
                    return img
                except Exception as e:
                    self.logger.Error ('Could not retrieve url {}: {}'.format(url,e))
                    return None
            else:
                self.more_images_to_read = False
                self.next_frameid_to_read = 0
                return None 



from pyzm.helpers.Base import Base
import pyzm.ml.object as  ObjectDetect
import pyzm.ml.face as FaceDetect
import re
import datetime
from pyzm.helpers.Media import MediaStream
import cv2


class DetectSequence(Base):
    def __init__(self, logger=None, options={}):
        super().__init__(logger)
        self.sequence = options.get('sequence', 'object').split(',')
        self.object_model = None
        self.face_model = None
        self.alpr_model = None
        self.ml_options = options

    def load_models(self):
        for seq in self.sequence:
            if seq == 'object':
                if not self.object_model:
                    self.object_model = ObjDetect(options=self.ml_options.get(seq, {}))
            elif seq == 'face':
                if not self.face_model:
                    self.face_model = FaceDetect(options = self.ml_options.get(seq, {}))
            else: 
                self.logger.Error('Invalid model {}'.format(seq))

    def detect_stream(self, stream, options={}):
        """Implements detection on a video stream

        Args:
            stream (string): location of media (file, url or event ID)
            api (object): instance of the API if the stream need to route via ZM
            options (dict, optional): Various options that control the detection process. Defaults to {}:
            - 'download': boolean # if True, will download video before analysis
            - 'start_frame' int # Which frame to start analysis. Default 1.
            - 'frame_skip': int # Number of frames to skip in video (example, 3 means process every 3rd frame)
            - 'max_frames' : int # Total number of frames to process before stopping
            - 'pattern': string # regexp for objects that will be matched. 'strategy' key below will be applied
                         to only objects that match this pattern
            - 'frame_set': string with exact frames to use (comma separated)
            - 'strategy': string # various conditions to stop matching as below
                'first' # stop at first match (Default)
                'most' # match the frame that has the highest number of detected objects
                'most_unique' # match the frame that has the highest number of unique detected objects
            - 'resize': int # width to resize image, default 800
            Returns:
                boxes (array): list of bounding boxes for matched frame
                labels (array): list of labels for matched frame
                confidences (array): list of confidences for matched frame
                id (int): frame id of matched frame
                img (cv2 image): image grab of matched frame
                all_matches (array of objects): list of boxes,labels,confidences of all frames matched

            Note:
            The same frames are not retrieved depending on whether you set
            ``download`` to ``True`` or ``False``. When set to ``True``, we use
            OpenCV's frame reading logic and when ``False`` we use ZoneMinder's image.php function
            which uses time based approximation. Therefore, the retrieve different frame offsets, but I assume
            they should be reasonably close.
            
        """

        

        stream_options = options
        match_strategy = options.get('strategy', 'first')
        match_pattern = re.compile(options.get('pattern', '.*'))
        all_matches = []
        matched_b = []
        matched_l = []
        matched_c = []
        matched_frame_id = None
        matched_frame_img = None
        manual_locking = False

        if len(self.sequence) > 1:
            manual_locking = False
            self.logger.Debug(1,'Using automatic locking as we are switching between models')
        else:
            manual_locking = True
            self.logger.Debug(1,'Using manual locking as we are only using one model')
            for seq in self.sequence:
                self.ml_options[seq]['auto_lock'] = False
            

        start = datetime.datetime.now()
        media = MediaStream(stream,'video', stream_options )
        while media.more():
            frame = media.read()
            if frame is None:
                self.logger.Debug(1,'Ran out of frames to read')
                break
            #fname = '/tmp/{}.jpg'.format(media.get_last_read_frame())
            #print (f'Writing to {fname}')
            #cv2.imwrite( fname ,frame)
            b = []
            l = []
            c = []
            for seq in self.sequence:
                self.logger.Debug(3,'Running sequence: {}'.format(seq))
                if seq == 'object':
                    if not self.object_model:
                        self.object_model = ObjectDetect.Object(options=self.ml_options.get(seq, {}))
                        if manual_locking:
                            self.object_model.acquire_lock()
                    _b,_l,_c, = self.object_model.detect(image=frame)
                    if (len(_l)):
                        b.append(_b)
                        l.append(_l)
                        c.append(_c)
                elif seq == 'face':
                    if not self.face_model:
                        self.face_model = FaceDetect.Face(options = self.ml_options.get(seq, {}))
                        if manual_locking:
                            self.face_model.acquire_lock()
                    _b,_l,_c, = self.face_model.detect(image=frame)
                    if (len(_l)):
                        b.extend(_b)
                        l.extend(_l)
                        c.extend(_c)
                  #  print (f'FACE: {b} {l} {c}')

           
                #print ('LABELS {} BOXES {}'.format(l,b))
                f_b = []
                f_l = []
                f_c = []

                match = list(filter(match_pattern.match, _l))
                for idx, label in enumerate(_l):
                    if label not in match:
                        continue
                    f_b.append(_b[idx])
                    f_l.append(label)
                    f_c.append(_c[idx])
                l = f_l
                b = f_b
                c = f_c

                if not len(l):
                    continue
                #print ('Frame:{}, labels:{}'.format(frames_processed, l))
                all_matches.append (
                    {
                        'frame_id': media.get_last_read_frame(),
                        'boxes': b,
                        'labels': l,
                        'confidences': c
                    }
                )
                if  ((match_strategy == 'first') 
                    or ((match_strategy == 'most') and (len(l) > len(matched_l))) 
                    or ((match_strategy == 'most_unique') and (len(set(l)) > len(set(matched_l))))):
                    matched_b = b
                    matched_c = c
                    matched_l = l
                    matched_frame_id = media.get_last_read_frame()
                    matched_frame_img = frame
                
                if (match_strategy=='first'):
                    break
            
        # release resources
        diff_time = (datetime.datetime.now() - start).microseconds / 1000

       
        if manual_locking:
            if self.face_model:
                self.face_model.release_lock()
            if self.object_model:
                self.object_model.release_lock()
        self.logger.Debug(
            1,'detection (with image loads) took: {} milliseconds to process {}'.format(diff_time, stream))
        media.stop()
        return matched_b, matched_l, matched_c, matched_frame_id, matched_frame_img, all_matches


        


"""
ZMLog
=======
Implements a python implementation of ZoneMinder's logging system
You could use this to connect it to the APIs if you want

"""

from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import select
from sqlalchemy import or_
from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError

import configparser
import glob,os,psutil
import syslog
from inspect import getframeinfo,stack
import time
import pwd,grp
import datetime


logger = None
pid = None
process_name = None
inited = False
config={}
engine = None
conn = None
connected = False
config_table = None
log_table = None
meta = None
log_fname = None
log_fhandle = None


connected = False
levels = {
    'DBG':1,
    'INF':0,
    'WAR':-1,
    'ERR':-2,
    'FAT':-3,
    'PNC':-4,
    'OFF':-5
    }

priorities = {
        'DBG':syslog.LOG_DEBUG,
        'INF':syslog.LOG_INFO,
        'WAR':syslog.LOG_WARNING,
        'ERR':syslog.LOG_ERR,
        'FAT':syslog.LOG_ERR,
        'PNC':syslog.LOG_ERR
    }

def init(name=None, override={}):
    """Initializes the ZM logging system. It follows ZM logging principles and ties into appropriate ZM logging levels. Like the rest of ZM, it can write to files, syslog and the ZM DB.

    To make it simpler to override, you can pass various options in the override dict. When passed, they will override any ZM setting
    
    Args:
        name (string, optional): Name to be used while writing logs. If not specified, it will use the process name. Defaults to None.
        override (dict, optional): Various parameters that can supercede standard ZM logs. Defaults to {}. The parameters that can be overriden are::

            {
                'dump_console': False,
                'conf_path': '/etc/zm',
                'user' : None,
                'password' : None,
                'host' : None,
                'webuser': 'www-data',
                'webgroup': 'www-data',
                'dbname' : None,
                'logpath' : '/var/log',
                'log_level_syslog' : 0,
                'log_level_file' : 0,
                'log_level_db' : 0,
                'log_debug' : 0,
                'log_level_debug' : 1,
                'log_debug_target' : '',
                'log_debug_file' :'',
                'server_id': 0,
                'driver': 'mysql+mysqlconnector'
            }
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    inited = True
    pid =  os.getpid()
    process_name = name or psutil.Process(pid).name()
    syslog.openlog(logoption=syslog.LOG_PID)

    config = {
        'conf_path': '/etc/zm',
        'user' : None,
        'password' : None,
        'host' : None,
        'webuser': 'www-data',
        'webgroup': 'www-data',
        'dbname' : None,
        'logpath' : '/var/log',
        'log_level_syslog' : 0,
        'log_level_file' : 0,
        'log_level_db' : 0,
        'log_debug' : 0,
        'log_level_debug' : 1,
        'log_debug_target' : '',
        'log_debug_file' :'',
        'server_id': 0,
        'driver': 'mysql+mysqlconnector',
        'dump_console': False
    }

    # Round 1 of overrides, before we read params from DB
        # Override with locals if present
    for key in config:
        if key in override:
            config[key] = override[key]
    

    # read all config files in order
    files=[]
    for f in glob.glob(config['conf_path']+'/conf.d/*.conf'):
        files.append(f)
    files.sort()
    files.insert(0,config['conf_path']+'/zm.conf')
    config_file = configparser.ConfigParser(interpolation=None)
    for f in files:
        with open(f) as s:
            #print ('reading {}'.format(f))
            config_file.read_string('[zm_root]\n' + s.read())

    # config_file will now contained merged data
    conf_data=config_file['zm_root']

    config['user'] = conf_data.get('ZM_DB_USER', 'zmuser')
    config['password'] = conf_data.get('ZM_DB_PASS', 'zmpass')
    config['webuser'] = conf_data.get('ZM_WEB_USER', 'www-data')
    config['webgroup'] = conf_data.get('ZM_WEB_GROUP', 'www-data')
    config['host'] = conf_data.get('ZM_DB_HOST', 'localhost')
    config['dbname'] = conf_data.get('ZM_DB_NAME', 'zm')
    config['logpath'] =  config_file['zm_root']['ZM_PATH_LOGS']

    cstr = config['driver']+'://{}:{}@{}/{}'.format(config['user'],
        config['password'],config['host'],config['dbname'])

    try:
        engine = create_engine(cstr, pool_recycle=3600)
        conn = engine.connect()
        connected = True
    except SQLAlchemyError as e:
        connected = False
        conn = None
        engine = None
        syslog.syslog (syslog.LOG_ERR, _format_string("Turning DB logging off. Could not connect to DB, message was:" + str(e)))
        config['log_level_db'] = levels['OFF']
        
    else:
        meta = MetaData(engine,reflect=True)
        config_table = meta.tables['Config']
        log_table = meta.tables['Logs']

        select_st = select([config_table.c.Name, config_table.c.Value]).where(
                or_(config_table.c.Name=='ZM_LOG_LEVEL_SYSLOG',
                    config_table.c.Name=='ZM_LOG_LEVEL_FILE',
                    config_table.c.Name=='ZM_LOG_LEVEL_DATABASE',
                    config_table.c.Name=='ZM_LOG_DEBUG',
                    config_table.c.Name=='ZM_LOG_DEBUG_LEVEL',
                    config_table.c.Name=='ZM_LOG_DEBUG_FILE',
                    config_table.c.Name=='ZM_LOG_DEBUG_TARGET',
                    config_table.c.Name=='ZM_SERVER_ID',
                    ))
        resultproxy = conn.execute(select_st)
        db_vals = {row[0]:row[1] for row in resultproxy}
        config['log_level_syslog'] = int(db_vals['ZM_LOG_LEVEL_SYSLOG']) 
        config['log_level_file'] = int(db_vals['ZM_LOG_LEVEL_FILE'])
        config['log_level_db'] = int(db_vals['ZM_LOG_LEVEL_DATABASE'])
        config['log_debug'] = int(db_vals['ZM_LOG_DEBUG'])
        config['log_level_debug'] = int(db_vals['ZM_LOG_DEBUG_LEVEL'])
        config['log_debug_file'] = db_vals['ZM_LOG_DEBUG_FILE']
        config['log_debug_target'] = db_vals['ZM_LOG_DEBUG_TARGET']
        config['server_id'] = db_vals.get('ZM_SERVER_ID',0)
    # Round 2 of overrides, after DB data is read
    # Override with locals if present
    for key in config:
        if key in override:
            config[key] = override[key]

    log_fname = None
    log_fhandle = None

    if config['log_level_file'] > levels['OFF']:
        
        n = os.path.split(process_name)[1].split('.')[0]
        log_fname = config['logpath']+'/'+n+'.log' 
        try:
            log_fhandle = open (log_fname,'a')
            uid = pwd.getpwnam(config['webuser']).pw_uid
            gid = grp.getgrnam(config['webgroup']).gr_gid
            os.chown(log_fname, uid,gid)
        except OSError as e:
            syslog.syslog (syslog.LOG_ERR, _format_string("Error opening file log:" + str(e)))
            log_fhandle = None


def get_config():
    """Returns configuration of ZM logger
    
    Returns:
        dict: config params
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    return config

def reconnect():
    """Invoked by the logger if disconnection occurs
    
    Returns:
        boolean: True if reconnected
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    try:
        conn.close()
    except:
        pass

    try:
        engine = create_engine(cstr, pool_recycle=3600)
        conn = engine.connect()
        #inspector = inspect(engine)
        #print(inspector.get_columns('Config'))
        meta = MetaData(engine,reflect=True)
        config_table = meta.tables['Config']
        log_table = meta.tables['Logs']
        message = 'reconnecting to Database...'
        log_string = '{level} [{pname}] [{msg}]'.format(level='INF', pname=process_name, msg=message)
        syslog.syslog (syslog.LOG_INFO, log_string)
    except SQLAlchemyError as e:
        connected = False
        syslog.syslog (syslog.LOG_ERR, _format_string("Turning off DB logging due to error received:" + str(e)))
        return False
    else:
        connected = True
        return True
        

def close():
    """Closes all handles. Invoke this before you exit your app
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    if conn: conn.close()
    if engine: engine.dispose()
    syslog.closelog()
    if (log_fhandle): log_fhandle.close()

def _format_string(message='', level='ERR'):
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    log_string = '{level} [{pname}] [{message}]'.format(level=level, pname=process_name, message=message)
    return (log_string)

def _log(level='DBG', message='', caller=None, debug_level=1):
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    log_string=''
    if not inited:
        raise ValueError ('Logs not initialized')
    # first stack element will be the wrapper log function
    # second stack element will be the actual calling function
    #print (len(stack()))
    if not caller:
        idx = min(len(stack()), 2) #incase someone calls this directly
        caller = getframeinfo(stack()[idx][0])
        #print ('Called from {}:{}'.format(caller.filename, caller.lineno))

    # If we are debug logging, show level too
    disp_level=level
    if level=='DBG':
        disp_level = f'DBG{debug_level}'
    # write to syslog
   
    if levels[level] <= config['log_level_syslog']:
        log_string = '{level} [{pname}] [{msg}]'.format(level=disp_level, pname=process_name, msg=message)
        syslog.syslog (priorities[level], log_string)

    # write to db
    if levels[level] <= config['log_level_db']:
        if not connected:
            syslog.syslog (syslog.LOG_INFO, _format_string("Trying to reconnect"))
            if not reconnect():
                syslog.syslog (syslog.LOG_ERR, _format_string("reconnection failed, not writing to DB"))
            return False

        log_string = '{level} [{pname}] [{msg}]'.format(level=disp_level, pname=process_name, msg=message)
        component = process_name
        serverid = config['server_id']
        pid = pid
        l = levels[level]
        code = level
        line = caller.lineno

        try:
            cmd = log_table.insert().values(TimeKey=time.time(), Component=component, ServerId=serverid, Pid=pid, Level=l, Code=code, Message=message,File=os.path.split(caller.filename)[1], Line=line)
            conn.execute(cmd)
        except SQLAlchemyError as e:
            connected = False
            syslog.syslog (syslog.LOG_ERR, _format_string("Error writing to DB:" + str(e)))

    # write to file components
    if levels[level] <= config['log_level_file']:
        timestamp = datetime.datetime.now().strftime('%x %H:%M:%S')
        # 07/15/19 10:10:14.050651 zmc_m8[4128].INF-zm_monitor.cpp/2516 [Driveway: images:218900 - Capturing at 3.70 fps, capturing bandwidth 98350bytes/sec]
        fnfl ='{}:{}'.format(os.path.split(caller.filename)[1], caller.lineno)
        log_string = '{timestamp} {pname}[{pid}] {level} {fnfl} [{msg}]\n'.format(timestamp=timestamp, level=disp_level, pname=process_name, pid=pid, msg=message, fnfl=fnfl)
        if log_fhandle: 
            log_fhandle.write(log_string)
            log_fhandle.flush()

    if config['dump_console']:
        print (log_string)

def Info(message=None,caller=None):
    """Info level ZM message
    
    Args:
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    _log('INF',message,caller)

def Debug(level=1, message=None,caller=None):
    """Debug level ZM message
    
    Args:
        level (int): ZM Debug level
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """

    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    target = config['log_debug_target']
    if target:
        targets = [x.strip().lstrip('_') for x in target.split('|')]
        # if current name does not fall into debug targets don't log
        if not any(map(process_name.startswith, targets)):
            return

    
    if config['log_debug'] and level <= config['log_level_debug']:
        _log('DBG', message,caller, level)

def Warning(message=None,caller=None):
    """Warning level ZM message
    
    Args:
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    _log('WAR',message,caller)

def Error(message=None,caller=None):
    """Error level ZM message
    
    Args:
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    _log('ERR',message,caller)
    
def Fatal(message=None,caller=None):
    """Fatal level ZM message. Quits after logging
    
    Args:
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    _log('FAT',message,caller)
    close()
    exit(-1)

def Panic(message=None,caller=None):
    """Panic level ZM message. Quits after logging
    
    Args:
        message (string): Message to log
        caller (stack frame info, optional): Used to log caller id/line #. Picked up automatically if none. Defaults to None.
    """
    global logger, pid, process_name, inited, config, engine, conn, connected, levels, priorities, config_table, log_table, meta, log_fname, log_fhandle
    _log('PNC',message,caller)
    close()
    exit(-1)





"""
Monitors
=========
Holds a list of Monitors for a ZM configuration
Given monitors are fairly static, maintains a cache of monitors
which can be overriden 
"""


from pyzm.helpers.Monitor import Monitor
from pyzm.helpers.Base import Base
import requests

class Monitors(Base):
    def __init__(self,logger=None, api=None):
        super().__init__(logger)
        self.api = api
        self._load()

    def _load(self,options={}):
        self.logger.Debug(1,'Retrieving monitors via API')
        url = self.api.api_url +'/monitors.json'
        r = self.api._make_request(url=url)
        ms = r.get('monitors')
        self.monitors = []
        for m in ms:
           self.monitors.append(Monitor(monitor=m,api=self.api, logger=self.logger))


    def list(self):
        return self.monitors

    def add(self, options={}):
        """Adds a new monitor
        
        Args:
            options (dict): Set of attributes that define the monitor::

                {
                    'function': string # function of monitor
                    'name': string # name of monitor
                    'enabled': boolean
                    'protocol': string
                    'host': string
                    'port': int
                    'path': string
                    'width': int
                    'height': int
                    'raw': {
                        # Any other monitor value that is not exposed above. Example:
                        'Monitor[Colours]': '4',
                        'Monitor[Method]': 'simple'
                    }

                }
        
        Returns:
            json: json response of API request
        """
        url = self.api.api_url +'/monitors.json'
        payload = {}
        if options.get('function'):
            payload['Monitor[Function]'] = options.get('function')
        if options.get('name'):
            payload['Monitor[Name]'] = options.get('name')
        if options.get('enabled'):
            enabled = '1' if options.get('enabled') else '0'
            payload['Monitor[Enabled]'] = enabled
        if options.get('protocol'):
            payload['Monitor[Protocol]'] = options.get('protocol')
        if options.get('host'):
            payload['Monitor[Host]'] = options.get('host')
        if options.get('port'):
            payload['Monitor[Port]'] = str(options.get('port'))
        if options.get('path'):
            payload['Monitor[Path]'] = options.get('path')
        if options.get('width'):
            payload['Monitor[Width]'] = str(options.get('width'))
        if options.get('height'):
            payload['Monitor[Height]'] = str(options.get('height'))

        if options.get('raw'):
            for k in options.get('raw'):
                payload[k] = options.get('raw')[k]
               
        if payload:
            return self.api._make_request(url=url, payload=payload, type='post')
    
    def find(self, id=None, name=None):
        """Given an id or name, returns matching monitor object
        
        Args:
            id (int, optional): MonitorId of monitor. Defaults to None.
            name (string, optional): Monitor name of monitor. Defaults to None.
        
        Returns:
            :class:`pyzm.helpers.Monitor`: Matching monitor object
        """
        if not id and not name:
            return None
        match = None
        if id:
            key = 'Id'
        else:
            key = 'Name'
    
        for mon in self.monitors:
            if id and mon.id() == id:
                match = mon
                break
            if name and mon.name().lower() == name.lower():
                match = mon
                break
        return match
        


    




#!/usr/bin/python3

import io
import os
import re
import codecs

from setuptools import setup

#Package meta-data.
NAME = 'pyzm'
DESCRIPTION = 'ZoneMinder API, Logger and other base utilities for python programmers'
URL = 'https://github.com/pliablepixels/pyzm/'
AUTHOR_EMAIL = 'pliablepixels@gmail.com'
AUTHOR = 'Pliable Pixels'
LICENSE = 'GPL'
INSTALL_REQUIRES=[
    'psutil>=5.6.3', 
    'SQLAlchemy>=1.3.5', 
    'mysql-connector-python>=8.0.16', 
    'requests>=2.18.4', 
    'dateparser>=0.7.1',
    'websocket-client>=0.56.0',
    'progressbar33',
    'portalocker>=2.0.0',
    'imutils'
    ]


here = os.path.abspath(os.path.dirname(__file__))
# read the contents of your README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(name = NAME,
      python_requires='>=3.0.0',
      version = find_version('pyzm','__init__.py'),
      description = DESCRIPTION,
      long_description = long_description,
      long_description_content_type='text/markdown',
      author = AUTHOR,
      author_email = AUTHOR_EMAIL,
      url = URL,
      license = LICENSE,
      install_requires=INSTALL_REQUIRES,
      py_modules = [
                    'pyzm.api',
                    'pyzm.helpers.Base',
                    'pyzm.helpers.Monitors',
                    'pyzm.helpers.Monitor',
                    'pyzm.helpers.Events',
                    'pyzm.helpers.Event',
                    'pyzm.helpers.States',
                    'pyzm.helpers.State',
                    'pyzm.helpers.Configs',
                    'pyzm.helpers.Media',
                    'pyzm.helpers.utils',

                    'pyzm.ZMLog',
                    'pyzm.ZMEventNotification',
                    'pyzm.ZMMemory',

                    'pyzm.ml.alpr',
                    'pyzm.ml.face',
                    'pyzm.ml.face_train',
                    'pyzm.ml.object',
                    'pyzm.ml.coral_edgetpu',
                    'pyzm.ml.hog',
                    'pyzm.ml.yolo',
                    'pyzm.ml.detect_sequence'
                    ]
      )





import cv2
import argparse
import pickle
from sklearn import neighbors
import imutils
import math
import ssl
import os
import datetime
import face_recognition

from pyzm.helpers.Base import Base


class FaceTrain (Base):

    def __init__(self,logger=None, options={}):
        super().__init__(logger)
        self.options = options
        start = datetime.datetime.now()
        
        diff_time = (datetime.datetime.now() - start).microseconds / 1000
        self.logger.Debug(
            1,'Face Recognition library load time took: {} milliseconds'.format(
                diff_time))


    def train(self):
        start = datetime.datetime.now()
        known_images_path = self.options.get('known_images_path')
        train_model = self.options.get('face_train_model')
        knn_algo = self.options.get('face_recog_knn_algo')
        upsample_times = self.options.get('face_upsample_times')
        num_jitters = self.options.get('face_num_jitters')

        encoding_file_name = known_images_path + '/faces.dat'
        try:
            if (os.path.isfile(known_images_path + '/faces.pickle')):
                # old version, we no longer want it. begone
                self.logger.Debug(
                    1,'removing old faces.pickle, we have moved to clustering')
                os.remove(known_images_path + '/faces.pickle')
        except Exception as e:
            self.logger.Error('Error deleting old pickle file: {}'.format(e))

        directory = known_images_path
        ext = ['.jpg', '.jpeg', '.png', '.gif']
        known_face_encodings = []
        known_face_names = []

        try:
            for entry in os.listdir(directory):
                if os.path.isdir(directory + '/' + entry):
                    # multiple images for this person,
                    # so we need to iterate that subdir
                    self.logger.Debug(
                        1,'{} is a directory. Processing all images inside it'.
                        format(entry))
                    person_dir = os.listdir(directory + '/' + entry)
                    for person in person_dir:
                        if person.endswith(tuple(ext)):
                            self.logger.Debug(1,'loading face from  {}/{}'.format(
                                entry, person))

                            # imread seems to do a better job of color space conversion and orientation
                            known_face = cv2.imread('{}/{}/{}'.format(
                                directory, entry, person))
                            if known_face is None or known_face.size == 0:
                                self.logger.Error('Error reading file, skipping')
                                continue
                            #known_face = face_recognition.load_image_file('{}/{}/{}'.format(directory,entry, person))

                            face_locations = face_recognition.face_locations(
                                known_face,
                                model=train_model,
                                number_of_times_to_upsample=upsample_times)
                            if len(face_locations) != 1:
                                self.logger.Error(
                                    'File {} has {} faces, cannot use for training. We need exactly 1 face. If you think you have only 1 face try using "cnn" for training mode. Ignoring...'
                                .format(person, len(face_locations)))
                            else:
                                face_encodings = face_recognition.face_encodings(
                                    known_face,
                                    known_face_locations=face_locations,
                                    num_jitters=num_jitters)
                                known_face_encodings.append(face_encodings[0])
                                known_face_names.append(entry)
                                #self.logger.Debug ('Adding image:{} as known person: {}'.format(person, person_dir))

                elif entry.endswith(tuple(ext)):
                    # this was old style. Lets still support it. The image is a single file with no directory
                    self.logger.Debug(1,'loading face from  {}'.format(entry))
                    #known_face = cv2.imread('{}/{}/{}'.format(directory,entry, person))
                    known_face = cv2.imread('{}/{}'.format(directory, entry))
                    face_locations = face_recognition.face_locations(
                        known_face,
                        model=train_model,
                        number_of_times_to_upsample=upsample_times)

                    if len(face_locations) != 1:
                        self.logger.Error(
                                    'File {} has {} faces, cannot use for training. We need exactly 1 face. If you think you have only 1 face try using "cnn" for training mode. Ignoring...'
                                    .format(person), len(face_locations))
                    else:
                        face_encodings = face_recognition.face_encodings(
                            known_face,
                            known_face_locations=face_locations,
                            num_jitters=num_jitters)
                        known_face_encodings.append(face_encodings[0])
                        known_face_names.append(os.path.splitext(entry)[0])

        except Exception as e:
            self.logger.Error('Error initializing face recognition: {}'.format(e))
            raise ValueError(
                'Error opening known faces directory. Is the path correct?')

        # Now we've finished iterating all files/dirs
        # lets create the svm
        if not len(known_face_names):
            self.logger.Error(
                'No known faces found to train, encoding file not created')
        else:
            n_neighbors = int(round(math.sqrt(len(known_face_names))))
            self.logger.Debug(2,'Using n_neighbors to be: {}'.format(n_neighbors))
            knn = neighbors.KNeighborsClassifier(n_neighbors=n_neighbors,
                                                algorithm=knn_algo,
                                                weights='distance')

            self.logger.Debug(1,'Training model ...')
            knn.fit(known_face_encodings, known_face_names)

            f = open(encoding_file_name, "wb")
            pickle.dump(knn, f)
            f.close()
            self.logger.Debug(1,'wrote encoding file: {}'.format(encoding_file_name))
        diff_time = (datetime.datetime.now() - start).microseconds / 1000
        self.logger.Debug(
            1,'Face Recognition training took: {} milliseconds'.format(diff_time))
